from glob import glob


def refactor(string: str) -> str:
    return string.replace("/", ".").replace("\\", ".").replace(".py", "")


pytest_plugins = [
    refactor(fixture) for fixture in glob("tests/fixtures/*.py") if "__" not in fixture
]


from mlplatform_lib.dataclass import TableColumnInfo
from dataclasses import dataclass
from typing import List

# "colInfoList": [
#     {
#         "name": "review",
#         "type": "VARCHAR",
#         "alias": "review",
#         "nullable": true,
#         "primaryKey": false,
#         "foreignKey": false,
#         "description": ""
#     },
#     {
#         "name": "sentiment",
#         "type": "VARCHAR",
#         "alias": "sentiment",
#         "nullable": true,
#         "primaryKey": false,
#         "foreignKey": false,
#         "description": ""
#     },
#     {
#         "name": "id",
#         "type": "VARCHAR",
#         "alias": "id",
#         "nullable": true,
#         "primaryKey": false,
#         "foreignKey": false,
#         "description": ""
#     }
# ],
# "numRows": 0


@dataclass
class TableDescInfo:
    colInfoList: List[TableColumnInfo]
    numRows: int



from mlplatform_lib.hyperdata_client import HyperdataHttpClient, HyperdataUserAuth
import os
import pathlib
from typing import List
import pandas as pd


class HyperdataApi:
    def __init__(self):
        self.client = HyperdataHttpClient(hd_addr=os.environ["hyperdata_addr"])
        self.auth = HyperdataUserAuth(
            project_id=os.environ["project_id"],
            user_id=os.environ["user_id"],
            token=os.environ["token"],
        )
        if "serving_mount_path" in os.environ:
            # shutil.rmtree(os.environ["mount_path"])
            os.makedirs(os.path.dirname(os.environ["mount_path"]), exist_ok=True)
            os.symlink(os.environ["serving_mount_path"], os.environ["mount_path"])
        else:
            os.makedirs(self.get_experiment_path(), exist_ok=True)
            os.makedirs(self.get_workflow_path(), exist_ok=True)
            if "inference_id" in os.environ:
                os.makedirs(self.get_inference_path(), exist_ok=True)

    def download_train_csv(self) -> str:
        return self._download_csv(
            do_id=int(os.environ["do_id"]), data_rootpath=self.get_experiment_path(),
        )

    def download_retrain_csv(self) -> str:
        return self.download_train_csv()

    def download_inference_csv(self) -> str:
        return self._download_csv(
            do_id=int(os.environ["do_id"]), data_rootpath=self.get_inference_path(),
        )

    def get_workflow_id(self) -> int:
        return int(os.environ["workflow_id"])

    def get_inference_id(self) -> int:
        return int(os.environ["inference_id"])

    def _download_csv(self, do_id: int, data_rootpath: str) -> str:
        result, sep, line_delim = self.client.download_do_to_csv(self.auth, do_id)

        pathlib.Path(data_rootpath).mkdir(parents=True, exist_ok=True)
        csv_path = os.path.join(data_rootpath, "%d.csv" % do_id)

        csv_str = result.data.decode("utf-8")
        if line_delim != "\n":
            csv_str.replace(line_delim, "\n")
        with open(csv_path, "w") as f:
            f.write(csv_str)

        if sep != ",":
            data = pd.read_csv(csv_path, sep=sep)
            data.to_csv(csv_path, sep=",", index=False)

        return csv_path

    def get_experiment_path(self) -> str:
        return os.path.join(os.environ["mount_path"], os.environ["experiment_id"])

    def get_workflow_path(self) -> str:
        return os.path.join(
            os.environ["mount_path"], os.environ["experiment_id"], os.environ["workflow_id"]
        )

    def get_prev_workflow_paths(self) -> List[str]:
        return [
            os.path.join(self.get_experiment_path(), str(workflow_id))
            for workflow_id in range(1, self.get_workflow_id())
        ]

    def get_inference_path(self):
        return os.path.join(
            os.environ["mount_path"],
            os.environ["experiment_id"],
            os.environ["workflow_id"],
            os.environ["inference_id"],
        )



from mlplatform_lib.hyperdata_client import HyperdataUserAuth
from mlplatform_lib.pre_check_base import PreCheckBase
import yaml


class PreCheckPredefinedAI(PreCheckBase):
    def __init__(self, server_config_path: str):
        super().__init__(server_config_path)
        server_config = yaml.safe_load(open(server_config_path))

        try:
            self.train_do_name = server_config["train_do_name"]
            self.inference_do_name = server_config["inference_do_name"]
        except KeyError as e:
            print(f'Cannot found "{str(e)}" in server_config.yaml')
            exit(1)

    def _check_train_do_exist(self):
        self._print_header(f'check is train do "{self.train_do_name}" exist.')
        self.auth = HyperdataUserAuth(
            project_id=self.mlplatform_request_header["projectId"],
            authorization=self.mlplatform_request_header["Authorization"],
            user_id=self.mlplatform_request_header["userId"],
        )
        do_list = self.hyperdata_client.get_do_list(self.auth)
        for do_info in do_list:
            if do_info.name == self.train_do_name:
                self.train_do_id = do_info.id
                print("check train do end.\n")
                return

        print("cannot found train do")
        exit(1)

    def _check_inference_do_exist(self):
        self._print_header(f'check is inference do "{self.inference_do_name}" exist.')
        do_list = self.hyperdata_client.get_do_list(self.auth)

        for do_info in do_list:
            if do_info.name == self.inference_do_name:
                print("check inference do end.\n")
                self.inference_do_id = do_info.id
                return

        print("cannot found inference do")
        exit(1)

    def run(self):
        super().run()
        self._check_train_do_exist()
        self._check_inference_do_exist()



from mlplatform_lib.hyperdata_api import HyperdataApi

import argparse
import json
import os
import pandas as pd
import requests
import sys
from typing import Dict
import yaml


class PredefinedAIApi(HyperdataApi):
    def __init__(self):
        self.hd_api = HyperdataApi()
        if "webserver_addr" not in os.environ:
            self.is_local = True
        else:
            self.is_local = False

            self.server_addr = os.path.join(os.environ["webserver_addr"], "predefinedai", "v2")
            self.headers = {
                "Content-Type": "application/json",
                "Project-Id": os.environ["project_id"],
                "Token": os.environ["token"],
                "User-Id": os.environ["user_id"],
            }

    def get_model_infos(self):
        if not self.is_local:
            res = requests.get(
                os.path.join(
                    self.server_addr, "model", "?experiment_id=" + str(os.environ["experiment_id"]),
                ),
                headers=self.headers,
                verify=False,
            )
            return json.loads(res.text)
        else:
            print("Current mode is local, Skip get_model_infos.")

    def insert_model_info(
        self, model_name: str, algorithm: str, metric: str, metric_result: str, model_json: Dict,
    ) -> int:
        if not self.is_local:
            input_json = {
                "name": model_name,
                "algorithm": algorithm,
                "metric": metric,
                "metricResult": metric_result,
                "modelJson": json.dumps(model_json),
            }
            res = requests.post(
                os.path.join(
                    self.server_addr,
                    "experiments",
                    str(os.environ["experiment_id"]),
                    "trains",
                    str(os.environ["train_id"]),
                    "models"
                ),
                headers=self.headers,
                data=json.dumps(input_json),
                verify=False,
            )
            if res.status_code != 200:
                raise Exception(f"status code {res.status_code} failed. {res.text}")
            model_dict = json.loads(res.text)
            return int(model_dict["id"])
        else:
            print("Current mode is local, Skip insert_model_info.")
            return 1

    def insert_visualizations(self, model_id: int, type: str, result: str) -> None:
        if not self.is_local:
            input_json = {
                "experiment_id": os.environ["experiment_id"],
                "model_id": model_id,
                "type": type,
                "result": result,
            }
            res = requests.post(
                os.path.join(self.server_addr, "visualization"),
                headers=self.headers,
                data=json.dumps(input_json),
                verify=False,
            )
            if res.status_code != 200:
                raise Exception(f"status code {res.status_code} failed. {res.text}")
        else:
            print("Current mode is local, Skip insert_visualizations.")

    def get_inference_csv_path(self):
        return os.path.join(self.hd_api.get_inference_path(), "inference.csv")

    def upload_inference_csv(self):
        if not self.is_local:
            if int(os.environ["target_do_id"]) == 0:
                print("target do id not selected. skip upload inference csv")
                return

            df = pd.read_csv(self.get_inference_csv_path())
            df_json = df.to_json(orient="split")

            inf_data = {
                "columns": df_json["columns"],
                "data": df_json["data"],
                "experiment_id": os.environ["experiment_id"],
                "target_do_id": os.environ["target_do_id"],
                "is_truncated": os.environ["is_truncated"],
            }
            res = requests.post(
                url=os.path.join(self.server_addr, "inference"),
                headers=self.headers,
                data=json.dumps(inf_data),
                verify=False,
            )
            if res.status_code != 200:
                raise Exception(f"status code {res.status_code} failed. {res.text}")
        else:
            print("Current mode is local, Skip upload_inference_csv.")
            return ""


class PredefinedAIArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.predefinedai_args = {}
        self.predefinedai_support_args = (
            "PredefinedAIArgumentParser.int",
            "PredefinedAIArgumentParser.float",
            "PredefinedAIArgumentParser.str",
            "PredefinedAIArgumentParser.column",
            "PredefinedAIArgumentParser.columns",
            "PredefinedAIArgumentParser.bool",
        )
        self.predefinedai_inner_func = (
            "int_parse",
            "float_parse",
            "str_parse",
            "column_parse",
            "columns_parse",
            "bool_parse",
        )

    def add_argument(self, *args, **kwargs):
        if "action" in kwargs and kwargs["action"] == "help":
            super().add_argument(*args, **kwargs)
            return

        if "type" not in kwargs or not kwargs["type"][0].__name__ in self.predefinedai_inner_func:
            raise ValueError(f"type must be in {self.predefinedai_support_args}")

        if "action" in kwargs:
            raise ValueError("Cannot use action in PredefinedAIArgumentParser.")

        parse_func, local_params = kwargs["type"]
        kwargs["type"] = parse_func
        super().add_argument(*args, **kwargs)

        if not args[0].startswith("--"):
            raise ValueError("name must start with --")

        name = args[0][2:]
        default = None
        if "default" in kwargs:
            default = kwargs["default"]

        description = None
        if "help" in kwargs:
            description = kwargs["help"]
        else:
            description = ""

        required = None
        if "required" in kwargs:
            required = kwargs["required"]
        else:
            required = False

        choices = None
        if "choices" in kwargs:
            choices = kwargs["choices"]

        arg_dict = {"name": name, "default": default, "description": description, "required": required}
        if parse_func.__name__ == "int_parse":
            if default is None:
                default = local_params["max"] - local_params["min"]

            if choices is None and local_params["window"] == 0:
                raise ValueError("Please specify choices or window.")

            if choices is not None and local_params["window"] != 0:
                raise ValueError("Cannot use choices and window both.")

            arg_dict.update({"type": "int", "min": local_params["min"], "max": local_params["max"]})
            if choices is not None:
                arg_dict["choices"] = choices
            else:
                arg_dict["window"] = local_params["window"]
        elif parse_func.__name__ == "float_parse":
            if default is None:
                default = local_params["max"] - local_params["min"]

            if choices is None and local_params["window"] == 0.0:
                raise ValueError("Please specify choices or window.")

            if choices is not None and local_params["window"] != 0.0:
                raise ValueError("Cannot use choices and window both.")

            arg_dict.update({"type": "float", "min": local_params["min"], "max": local_params["max"]})
            if choices is not None:
                arg_dict["choices"] = choices
            else:
                arg_dict["window"] = local_params["window"]
        elif parse_func.__name__ == "str_parse":
            if default is None:
                default = ""

            arg_dict.update({"type": "str"})
            if choices is not None:
                arg_dict["choices"] = choices
        elif parse_func.__name__ == "column_parse":
            if default is None:
                default = ""

            arg_dict.update({"type": "column"})
        elif parse_func.__name__ == "columns_parse":
            if default is None:
                default = "[]"

            arg_dict.update(
                {
                    "type": "columns",
                    "name": name,
                    "default": default,
                    "description": description,
                    "required": required,
                }
            )
        elif parse_func.__name__ == "bool_parse":
            if default is None:
                default = True

            arg_dict.update({"type": "bool"})
        else:
            raise ValueError(f"parse function {parse_func.__name__} cannot found.")

        self.predefinedai_args[arg_dict["name"]] = arg_dict

    @staticmethod
    def int(min: int = 0, max: int = sys.maxsize, window: int = 1):
        def int_parse(arg):
            try:
                val = int(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to integer.")

            if min < max:
                raise ValueError(f"{min} must be less than {min}")
            elif val < min:
                raise ValueError(f"{arg} must be greater than {min}.")
            elif val > max:
                raise ValueError(f"{arg} must be less than {max}.")
            elif (val - min) % window != 0:
                raise ValueError(f"'({arg} - {min}) mod {window}' must be zero.")
            return val

        return int_parse, locals()

    @staticmethod
    def float(min: float = 0.0, max: float = sys.float_info.max, window: float = 1.0):
        def float_parse(arg):
            try:
                val = float(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to float.")

            if val < min:
                raise ValueError(f"{arg} must be greater than {min}.")
            elif val > max:
                raise ValueError(f"{arg} must be less than {max}.")
            elif (val - min) % window != 0.0:
                raise ValueError(f"'({arg} - {min}) mod {window}' must be zero.")
            return val

        return float_parse, locals()

    @staticmethod
    def str():
        def str_parse(arg):
            try:
                val = str(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to str.")

            return val

        return str_parse, locals()

    @staticmethod
    def column():
        def column_parse(arg):
            try:
                val = str(arg)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{arg} cannot cast to str. column must be str.")

            return val

        return column_parse, locals()

    @staticmethod
    def columns():
        def columns_parse(arg):
            try:
                val = json.loads(arg)
            except Exception as e:
                print(e)
                raise argparse.ArgumentTypeError(
                    f"{arg} cannot cast to dictionary. columns must be json string."
                )

            return val

        return columns_parse, locals()

    @staticmethod
    def bool():
        def bool_parse(arg):
            if isinstance(arg, bool):
                return arg
            if arg.lower() in ("yes", "true", "t", "y", "1"):
                return True
            elif arg.lower() in ("no", "false", "f", "n", "0"):
                return False
            else:
                raise argparse.ArgumentTypeError("Boolean value expected.")

        return bool_parse, locals()

    def update_train_config(self, base_dir="../config"):
        if not os.path.isdir(base_dir):
            os.makedirs(base_dir, exist_ok=True)

        train_config_path = os.path.join(base_dir, "train_config.yaml")
        args_list = [val for key, val in self.predefinedai_args.items()]
        if not os.path.isfile(train_config_path):
            train_dict = {
                "train": {
                    "model": "",
                    "pipeline_image_info": {
                        "name": "",
                        "task_type": "",
                        "image": "",
                        "image_pull_policy": "Always",
                        "cpu": "",
                        "memory": "",
                        "gpu": "",
                        "working_dir": "",
                        "args": {},
                        "envs": {},
                    },
                }
            }
            train_dict["train"]["pipeline_image_info"]["args"] = args_list
            with open(train_config_path, "w") as f:
                yaml.dump(train_dict, f)
        else:
            with open(os.path.join(train_config_path), "r+") as f:
                train_dict = yaml.safe_load(f)
                train_dict["train"]["pipeline_image_info"]["args"] = args_list
                yaml.dump(train_dict, f)



from dataclasses import dataclass
from .column_of_table_from_default_source import ColumnOfTableFromDefaultSource
from typing import List


@dataclass
class TableFromDefaultDataSourceInfo:
    tableName: str
    columnList: List[ColumnOfTableFromDefaultSource]



from mlplatform_lib.hyperdata_client import HyperdataHttpClient
import yaml
import requests
import json


class PreCheckBase:
    def __init__(self, server_config_path: str):
        server_config = yaml.safe_load(open(server_config_path))
        try:
            self.hyperdata_addr = server_config["hyperdata_addr"]
            self.mlplatform_addr = server_config["mlplatform_addr"]
            self.proauth_addr = server_config["proauth_addr"]
            self.user_id = server_config["user_id"]
            self.user_password = server_config["user_password"]
            self.project_name = server_config["project_name"]
        except KeyError as e:
            print(f'Cannot found "{str(e)}" in server_config.yaml')
            exit(1)

        self.hyperdata_client = HyperdataHttpClient(self.hyperdata_addr)

    def _print_header(self, msg: str):
        print("-" * len(msg))
        print(msg)
        print("-" * len(msg))

    def _check_hyperdata_proauth_server_connect(self):
        self._print_header(f'check proauth server "{self.proauth_addr}" connection.')
        res = requests.post(
            url=self.proauth_addr + "/proauth/oauth/authenticate",
            headers={"Content-Type": "application/json;charset=UTF-8"},
            data=json.dumps({"user_id": self.user_id, "password": self.user_password}),
        )
        if res.status_code != 200:
            print(
                "cannot connect to %s. proauth server returns status code %d"
                % (self.proauth_addr, res.status_code)
            )
            raise Exception(res)

        response = json.loads(res.text)
        if response["result"] == "false":
            raise Exception(response["error"])
        self.access_token = response["token"]
        self.hyperdata_request_header = {"Authorization": self.access_token}
        print("check proauth server connection end.\n")

    def _check_hyperdata_project(self):
        self._print_header(f'check is hyperdata project "{self.project_name}" exist.')

        project_get_url = self.hyperdata_addr + "/hyperdata/web-service/projects"
        res = requests.get(project_get_url, headers=self.hyperdata_request_header)
        if res.status_code != 200:
            print(
                "cannot connect to %s. hyperdata server returns status code %d"
                % (self.hyperdata_addr, res.status_code)
            )
            raise Exception(res)

        project_infos = json.loads(res.text)["dto"]["project"]
        project_id = None
        for project_info in project_infos:
            if project_info["name"] == self.project_name:
                project_id = project_info["id"]
                break
        if project_id is None:
            print(f"Cannot found project {self.project_name}. Please check hyperdata.")
            raise Exception("_check_hyperdata_project failed.")
        else:
            self.project_id = project_id

        print("check hyperdata server connection end.\n")

    def _check_mlplatform_server_connect(self):
        self._print_header(
            f'check mlplatform server "{self.mlplatform_addr}" connection.'
        )
        self.mlplatform_request_header = {
            "projectId": str(self.project_id),
            "Authorization": self.access_token,
            "userId": self.user_id,
            "authorizationType": "hyperdata",
            "Content-Type": "application/json"
        }
        res = requests.get(self.mlplatform_addr, headers=self.mlplatform_request_header)
        if res.status_code != 200:
            print(
                "cannot connect to %s. mlplatform server returns status code %d"
                % (self.mlplatform_addr, res.status_code)
            )
            raise Exception(res)
        print("check mlplatform server connection end.\n")

    def run(self):
        self._check_hyperdata_proauth_server_connect()
        self._check_hyperdata_project()
        self._check_mlplatform_server_connect()



from setuptools import setup, find_packages

setup(
    name="hyperdata",
    version="8.4.0.0",
    description="hyperdata",
    author="hyperdata",
    author_email="mansu_kim@tmax.co.kr",
    url="",
    download_url="",
    install_requires=[],
    packages=find_packages(exclude=[]),
    keywords=["hyperdata"],
    python_requires=">=3.7",
    package_data={},
    zip_safe=False,
    setup_requires=["pandas>=1.0.1", "requests>=2.25.1"],
    classifiers=["Programming Language :: Python :: 3.7"],
)



from dataclasses import dataclass
# {
#     "name": "review",
#     "type": "VARCHAR",
#     "alias": "review",
#     "nullable": true,
#     "primaryKey": false,
#     "foreignKey": false,
#     "description": ""
# },


@dataclass
class TableColumnInfo:
    name: str
    type: str
    alias: str
    nullable: bool
    primaryKey: bool
    foreignKey: bool
    description: str



from mlplatform_lib.predefinedai.pre_check_predefinedai import PreCheckPredefinedAI

import json
import requests
from time import sleep
import uuid
import yaml


class PredefinedAITemplateTest:
    def __init__(self, server_config_path: str, image_config_path: str):
        # replace test params
        server_config = yaml.safe_load(open(server_config_path))
        self.pre_check = PreCheckPredefinedAI(server_config_path)
        self.pre_check.run()

        self.mlplatform_addr = server_config["mlplatform_addr"]
        self.image_config = yaml.safe_load(open(image_config_path))
        self.experiment_id = None

    def _print_train_logs(self, experiment_id, train_id):
        self.pre_check._print_header("train log start")
        log_url = f"{self.mlplatform_addr}/experiments/{experiment_id}/trains/{train_id}/logs/predefinedai"
        res = requests.get(log_url, headers=self.pre_check.mlplatform_request_header)
        logs = json.loads(res.text)
        for log in logs:
            print(f"##########{log['name']}##########")
            print(log["log"])
            print("\n")
    '''
    def _print_inference_logs(self, experiment_id, inference_id):
        self.pre_check._print_header("inference log start")
        log_url = self.mlplatform_addr + "/predefinedai/v2/log/inference"
        log_url += "?experiment_id=" + str(experiment_id)
        log_url += "&inference_id=" + str(inference_id)
        res = requests.get(log_url, headers=self.request_header)
        print(res.text)
        logs = json.loads(res.text)
        for log in logs:
            print(f"##########{log['name']}##########")
            print(log["log"])
            print("\n")
    '''

    def _experiment(self):
        self.pre_check._print_header("predefinedai template experiment start")

        experiment_url = f"{self.pre_check.mlplatform_addr}/experiments"
        model_name = self.image_config["train"]["model"]
        random_generated_experiment_name = (
            f"{model_name}_{str(uuid.uuid4())}"
        )

        create_experiment_data = {
            "name": random_generated_experiment_name,
            "experimentType": "predefinedai",
            "pvcSize": 10,
            "description": f"{random_generated_experiment_name} description.",
            "doId": self.pre_check.train_do_id,
            "taskType": model_name
        }

        print(f"experiment_name: {random_generated_experiment_name}")
        print(
            "send create experiment request to mlplatform server. After creating experiments,"
            + "training will start."
        )
        res = requests.post(
            experiment_url,
            json=create_experiment_data,
            headers=self.pre_check.mlplatform_request_header
        )
        if res.status_code != 200:
            raise Exception(f"create experiment failed.\nstatus_code: {res.status_code}\nreason: {res.text}")
        else:
            # set experiment id for retraining
            self.experiment_id = json.loads(res.text)["id"]
            print("create experiment succeeded.\n")

    def _train(self):
        self.pre_check._print_header("predefinedai template train start")

        train_url = f"{self.mlplatform_addr}/experiments/{self.experiment_id}/trains"
        create_train_data = {
            "experimentType": "predefinedai",
            "workflow": {
                "maxCpuSize": 1,
                "maxMemorySize": 1,
                "maxGpuSize": 0,
                "description": "workflow test"
            },
            "args": self.image_config["train"]["template"]["args"],
            "envs": self.image_config["train"]["template"]["envs"]
        }

        res = requests.post(
            train_url, json=create_train_data, headers=self.pre_check.mlplatform_request_header
        )
        if res.status_code != 200:
            raise Exception(f"create train failed. reason: {res.text}")
        else:
            print("wait until train finished.")

            train_id = res["id"]
            train_url_get = f"{train_url}/{train_id}"
            while True:
                res = requests.get(train_url_get, headers=self.pre_check.mlplatform_request_header)

                if res["workflow"]["status"] == "failed":
                    self._print_train_logs(self.experiment_id, train_id)
                    raise Exception("train failed.")
                elif res["workflow"]["status"] == "Succeeded":
                    break
                sleep(5)
            print("train succeeded.\n")

    '''
    # 재학습은 내부적으로 학습 시에 사용했던 DO와 같은 DO를 사용하도록 되어 있습니다.
    # 따라서 do_id를 전송하지 않습니다.
    def _retrain(self):
        self.pre_check._print_header("retrain start")
        experiment_url = self.mlplatform_addr + "/predefinedai/v2/experiments"
        retrain_experiment_data = self.image_config["retrain"]
        retrain_experiment_data["experiment_id"] = self.experiment_id
        # retrain_experiment_data = {
        #    "experiment_id": self.experiment_id,
        # }
        # 덮어쓰기 가능한 목록
        # retrain_experiment_data = {
        #    "experiment_id": self.experiment_id,
        #    "pipeline_image_info": {
        #        "cpu": "2",
        #        "memory": "2Gi",
        #        "gpu": "0",
        #        "args": {
        #            "seed": 6
        #        },
        #        "envs": {}
        #    }
        # }
        print("send retrain experiment request to automl server.")
        res = requests.put(
            experiment_url, json=retrain_experiment_data, headers=self.request_header
        )
        if res.status_code != 200:
            raise Exception(f"retrain failed. reason: {res.text}")
        else:
            print("wait until retrain finished.")

            experiment_url_get = (
                experiment_url + "?experiment_id=" + str(self.experiment_id)
            )
            while True:
                res = requests.get(experiment_url_get, headers=self.request_header)
                if res.status_code != 200:
                    raise Exception(f"Get experiment info failed. " f"reason: {res.text}",)

                experiment_info = json.loads(res.text)
                current_workflow_infos = experiment_info["workflows"][-1]
                if current_workflow_infos["status"] == "Failed":
                    self._print_train_logs(self.experiment_id, current_workflow_infos["id"])
                    raise Exception("retrain failed.")
                elif current_workflow_infos["status"] == "Succeeded":
                    # set workflow info for next inference
                    self.workflow_id = current_workflow_infos["id"]
                    break
                sleep(5)
            print("retrain succeeded.\n")

    def _inference(self):
        self.pre_check._print_header("inference start")
        inference_url = self.mlplatform_addr + "/predefinedai/v2/inferenceservice"
        inference_data = self.image_config["inference"]
        inference_data["experiment_id"] = self.experiment_id
        inference_data["workflow_id"] = self.workflow_id
        inference_data["input_do_id"] = self.inference_do_id
        inference_data["is_truncated"] = False
        # inference_data = {
        #    "experiment_id": self.experiment_id,
        #    "workflow_id": self.workflow_id,
        #    "input_do_id": self.inference_do_id,
        #    "is_truncated": True,
        # }
        # 덮어쓰기 가능한 목록
        # inference_data = {
        #    "experiment_id": self.experiment_id,
        #    "workflow_id": self.workflow_id,
        #    "input_do_id": self.inference_do_id,
        #    "is_truncated": True,
        #    "pipeline_image_info": {
        #        "cpu": "2",
        #        "memory": "2Gi",
        #        "gpu": "0",
        #        "args": {
        #            "seed": 7
        #        },
        #        "envs": {}
        #    }
        # }
        print("send inference request to automl server.")
        res = requests.post(inference_url, json=inference_data, headers=self.request_header)
        if res.status_code != 200:
            raise Exception(f"inference failed. reason: {res.text}")
        else:
            print("wait until inference finished.")

            self.inference_id = json.loads(res.text)["inference_id"]
            inference_url_get = (
                inference_url
                + "?experiment_id="
                + str(self.experiment_id)
                + "&inference_id="
                + str(self.inference_id)
            )
            while True:
                res = requests.get(inference_url_get, headers=self.request_header)
                if res.status_code != 200:
                    raise Exception(f"Get inference info failed. " f"reason: {res.text}")

                res_json = json.loads(res.text)
                inference_status = res_json["inference_status"]
                if inference_status == "Failed":
                    self._print_inference_logs(self.experiment_id, res_json["inference_id"])
                    raise Exception("inference failed.")
                elif inference_status == "Finished":
                    break
                sleep(5)
            print("inference succeeded.\n")

    def _inference_download(self):
        self.pre_check._print_header("inference download start")
        inference_download_url = self.mlplatform_addr + "/predefinedai/v2/inferenceservice"
        inference_download_url += "?action=Download"
        inference_download_url += "&experiment_id=" + str(self.experiment_id)
        inference_download_url += "&inference_id=" + str(self.inference_id)
        res = requests.get(inference_download_url, headers=self.request_header, stream=True)
        if res.status_code == 200:
            print("inference download succeeded.\n")
        else:
            raise Exception(f"inference download failed. reason: {res.text}")

    def _serving(self):
        self.pre_check._print_header("serving start")
        serving_url = self.mlplatform_addr + "/predefinedai/v2/serving"
        # kfserving works based on knative.
        # and knative uses docker image digest instead of tag
        # refer. https://knative.dev/docs/serving/tag-resolution/
        # https://docs.google.com/presentation/d/e/2PACX-1vTgyp2lGDsLr_bohx3Ym_2mrTcMoFfzzd6jocUXdmWQFdXydltnraDMoLxvEe6WY9pNPpUUvM-geJ-g/pub?resourcekey=0-FH5lN4C2sbURc_ds8XRHeA&slide=id.p
        # docker image digest can find using docker pull command.
        # --example--
        # user> docker pull 127.0.0.1:5000/hyperdata/sentimentanalysis_base:20210322_v1
        # 20210322_v1: Pulling from hyperdata/sentimentanalysis_base
        # Digest: sha256:40618baabd44864647460d43721dbf548a43e5b20406e66d58fafd11b3ac02c3
        # Status: Image is up to date for 127.0.0.1:5000/hyperdata/sentimentanalysis_base:20210322_v1
        # 127.0.0.1:5000/hyperdata/sentimentanalysis_base:2021-0322_v1
        serving_data = self.image_config["serving"]
        serving_data["experiment_id"] = self.experiment_id
        serving_data["workflow_id"] = self.workflow_id
        # serving_data = {
        #    "experiment_id": self.experiment_id,
        #    "workflow_id": self.workflow_id,
        # }
        # 덮어쓰기 가능한 목록
        # serving_data = {
        #    "experiment_id": self.experiment_id,
        #    "workflow_id": self.workflow_id,
        #    "serving_image_info": {
        #        "requests_cpu": "1",
        #        "limits_cpu": "2",
        #        "requests_memory": "1Gi",
        #        "limits_memory": "2Gi",
        #        "requests_gpu": "0",
        #        "limits_gpu": "0",
        #        "min_replicas": "1",
        #        "max_replicas": "1",
        #        "args": {},
        #        "envs": {},
        #    },
        # }
        print("send start serving server request to mlplatform server.")
        res = requests.post(serving_url, json=serving_data, headers=self.request_header)
        if res.status_code != 200:
            raise Exception(f"serving failed. reason: {res.text}")
        else:
            print("wait until serving server is start.")
            while True:
                serving_url_get = (
                    serving_url
                    + "?experiment_id="
                    + str(self.experiment_id)
                    + "&workflow_id="
                    + str(self.workflow_id)
                )
                res = requests.get(serving_url_get, headers=self.request_header)
                if res.status_code not in [200, 503, 404]:
                    raise Exception(f"Get serving info failed. " f"reason: {res.text}")

                res_json = json.loads(res.text)
                if res_json["serve_status"] == "Possible":
                    service_host_name = res_json["service_host_name"]
                    url = res_json["url"]

                    # ref. https://github.com/kubeflow/kfserving/tree/release-0.4#test-kfserving-installation
                    # kfserving request header
                    headers = {"Host": service_host_name}
                    # kfserving input

                    result = self.pre_check.hyperdata_client.get_do_samples(
                        self.pre_check.auth, self.inference_do_id, 10
                    )
                    result_data = json.loads(result.data["dto"]["tableString"])
                    response_data = {
                        "columns": list(result_data[0].keys()),
                        "rows": [],
                    }
                    for data_dict in result_data:
                        response_data["rows"].append(
                            list(map(str, list(data_dict.values())))
                        )

                    request_json = {"instances": []}
                    for row in response_data["rows"]:
                        instance = {}
                        for idx, val in enumerate(row):
                            instance[response_data["columns"][idx]] = val
                        request_json["instances"].append(instance)
                    print(
                        "In test, serving uses inference input data as serving input data."
                    )
                    print("serving_input_data: ", request_json)
                    try_cnt = 0
                    while try_cnt < 10:
                        res = requests.post(
                            url, headers=headers, data=json.dumps(request_json)
                        )
                        if res.status_code == 200:
                            print("serving_output_data: ", res.text)
                            print("serving succeeded.")
                            return
                        sleep(5)
                        try_cnt += 1
                    raise Exception(f"serving failed. reason: {res.text}")
                sleep(5)
    '''

    def _delete_experiment(self):
        print("unittest finished.")
        if self.experiment_id is not None:
            print("delete experiment.")
            experiment_url = f"{self.pre_check.mlplatform_addr}/experiments"
            delete_experiment_data = [self.experiment_id]
            res = requests.delete(
                experiment_url, json=delete_experiment_data, headers=self.pre_check.mlplatform_request_header
            )
            if res.status_code != 200:
                print(f"delete experiment failed. reason:{res.text}")

    def run(self):
        try:
            self._experiment()
            self._train()
            '''
            self._retrain()
            self._inference()
            self._inference_download()
            self._serving()
            '''
            if self.experiment_id is not None:
                self._delete_experiment()
        except Exception as e:
            if self.experiment_id is not None:
                self._delete_experiment()
            raise e



from .data_object import DataObject
from .insert_tuple_object import InsertTupleObject
from .column_of_table_from_default_source import ColumnOfTableFromDefaultSource
from .table_from_default_datasource_info import TableFromDefaultDataSourceInfo
from .table_column_info import TableColumnInfo
from .table_desc_info import TableDescInfo

__all__ = [
    "DataObject",
    "InsertTupleObject",
    "ColumnOfTableFromDefaultSource",
    "TableFromDefaultDataSourceInfo",
    "TableColumnInfo",
    "TableDescInfo"
]
